/*
 * Determine if a given point is inside a cylinder.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

typedef struct {
	double	x, y, z;
} Point;

typedef struct {
	Point	cc1, cc2;	/* cap centers */
	double	radius;
} Cylinder;

int	ParseArguments (int argc, char* argv[], Point* pt, Cylinder* cyl);

double	ProyectPointOnCenterLine (Cylinder* cyl, Point* pt);
double	DistanceFromCenterLine   (Cylinder* cyl, double tmin, Point* pt);

void	ParametricPointOnCenterLine (Cylinder* cyl, double t, Point* pt);

void	PointSubtract  (Point* a, Point* b, Point* result);
double	DotProduct     (Point* a, Point* b);

int	main (int argc, char* argv[])
{
	Cylinder cyl;
	Point	 pt;
	double	 tmin, dist;

	if (!ParseArguments (argc, argv, &pt, &cyl)) {
		fprintf (stderr, "Usage: cylinder px py pz ax ay az bx by bz r\n"
				 "       where (px,py,pz): given point\n"
				 "             (ax,ay,az): first cap center\n"
				 "             (bx,by,bz): second cap center\n"
				 "                      r: cylinder radius\n");
		return 1;
	}

	tmin = ProyectPointOnCenterLine (&cyl, &pt);
	dist = DistanceFromCenterLine (&cyl, tmin, &pt);

	printf ("\ntmin: %9.5f\ndist: %9.5f\n\n", tmin, dist);

	if (tmin < 0 || tmin > 1)
		printf ("Point outside - beyond the caps\n");
	else if (dist > cyl.radius)
		printf ("Point outside - farther than the radius\n");
	else
		printf ("Point inside or touching\n");

	return 0;
}

int	ParseArguments (int argc, char* argv[], Point* pt, Cylinder* cyl)
{
	if (argc < 1+3+3+3+1)
		return 0;

	pt->x = atof (*++argv);
	pt->y = atof (*++argv);
	pt->z = atof (*++argv);

	cyl->cc1.x = atof (*++argv);
	cyl->cc1.y = atof (*++argv);
	cyl->cc1.z = atof (*++argv);

	cyl->cc2.x = atof (*++argv);
	cyl->cc2.y = atof (*++argv);
	cyl->cc2.z = atof (*++argv);

	cyl->radius = atof (*++argv);

	return 1;
}

double	ProyectPointOnCenterLine (Cylinder* cyl, Point* pt)
{
	Point	range;

	PointSubtract (&cyl->cc2, &cyl->cc1, &range);

	return (DotProduct (pt, &range) - DotProduct (&cyl->cc1, &range))
	      / DotProduct (&range, &range);
}

double	DistanceFromCenterLine (Cylinder* cyl, double tmin, Point* pt)
{
	Point	linept, diff;

	ParametricPointOnCenterLine (cyl, tmin, &linept);

	PointSubtract (&linept, pt, &diff);

	return sqrt (DotProduct (&diff, &diff));
}

void	ParametricPointOnCenterLine (Cylinder* cyl, double t, Point* pt)
{
	Point	range;

	PointSubtract (&cyl->cc2, &cyl->cc1, &range);

	pt->x = cyl->cc1.x + t * range.x;
	pt->y = cyl->cc1.y + t * range.y;
	pt->z = cyl->cc1.z + t * range.z;
}

void	PointSubtract (Point* a, Point* b, Point* result)
{
	result->x = a->x - b->x;
	result->y = a->y - b->y;
	result->z = a->z - b->z;
}

double	DotProduct (Point* a, Point* b)
{
	return a->x * b->x
	     + a->y * b->y
	     + a->z * b->z;
}
